/*
 AT Mini Matrix Ctrl
 Version: 0.4
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 20/02/2013
 Last Modified: 11/05/2013
 
 Displays an animation on an 8x8 LED Matrix depending on the triggers set and the current time.
 
 */

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include "setup.c"

// ATmega48A/88A/168PA/328P Pin map
//
//        21065432
//        DDDCCCCC
//        PPPPPPPP
//
//        21098765
//        33322222            
//       /--------+
//  PD3 1|        |24 PC1
//  PD4 2|        |23 PC0
//  GND 3|        |22 ADC7
//  VCC 4|        |21 GND
//  GND 5|        |20 AREF
//  VCC 6|        |19 ADC6
//  PB6 7|        |18 AVCC
//  PB7 8|        |17 PB5
//       +--------+
//        91111111
//         0123456
//
//        PPPPPPPP
//        DDDBBBBB
//        56701234
//
//
// PC0 Row0		PD4 Col0
// PC2 Row1		PC1 Col1
// PB3 Row2		PB5 Col2
// PC3 Row3		PD7 Col3
// PD5 Row4		PB4 Col4
// PB2 Row5		PD6 Col5
// PD3 Row6		PD2 Col6
// PB0 Row7		PD1 Col7

volatile uint32_t unix_timestamp = 1367728200; // Current unix timestamp (change this to the current time)
byte trigger_location = 0; // Where in the triggers array we are up to
uint16_t trigger_delay = 0; // Keep track of the time passed since the LEDs were lit
int leds_location = 0; // Where in the leds animation array we are

// Triggers array listed in ascending order
// 0 - Start timestamp
// 1 - End timestamp
// 2 - Seconds to wait before re-displaying the animation (in multiples of 8, change the incrementing timer from 8 seconds to 1 second to have 1 second accuracy)
// 3 - End location in ledProgram array
// 4 - Start location in ledProgram array
// 5 - Speed of the animation (higher means is slower animation)
// 6 - Animation (0) or text (1) array
// You need to make sure that the "2 - Seconds to wait" is more than it takes your animation to display
uint32_t triggers[3][7] = {
	{1000, 1020, 8, 41, 28, 8, 1},
	{1030, 1050, 8, 128, 319, 10, 0},
	{1060, 999999, 8, 321, 455, 30, 0}
};

// 8x8 array of LEDs used as our buffer before displaying them
byte LEDs[8][8] = {
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0}
};

// 8x16 buffer LEDs for the text scrolling
byte bufferLEDs[8][16] = {
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
};

// LEDs animations array
prog_uchar ledProgram[] PROGMEM = {
	// Single firework
	16,2,128,0,2,64,0,8,
	8,128,2,0,0,130,0,16,
	16,2,128,0,2,64,0,8,
	8,128,2,0,0,130,0,16,
	16,2,128,0,2,64,0,8,
	0,0,0,16,0,0,0,0,
	0,0,0,16,0,0,0,0,
	0,0,0,16,0,0,0,0,
	0,0,0,16,0,0,0,16,
	0,0,16,0,0,0,16,16,
	0,16,0,0,0,16,16,255,
	16,0,0,0,16,16,255,255,
	0,16,0,0,16,16,255,255,
	0,0,16,0,16,16,255,255,
	0,0,0,16,16,16,255,255,
	0,0,0,0,16,16,255,255,
	
	// Multi firework
	0,0,0,0,1,0,4,1,
	0,0,0,0,8,66,17,4,
	0,128,32,136,1,40,4,1,
	0,64,128,32,8,66,17,4,
	0,128,32,136,1,40,4,1,
	0,64,128,32,8,66,17,4,
	0,128,32,136,0,40,0,0,
	0,64,128,32,0,0,4,0,
	0,0,64,0,16,0,4,0,
	0,64,0,16,0,4,0,0,
	64,0,16,0,4,0,0,0,
	0,16,0,4,0,0,0,84,
	16,0,4,0,0,0,84,84,
	0,4,0,0,0,84,84,255,
	4,0,0,0,84,84,255,255,
	0,4,0,0,84,84,255,255,
	16,0,4,0,84,84,255,255,
	0,16,0,4,84,84,255,255,
	64,0,16,0,84,84,255,255,
	0,64,0,16,84,84,255,255,
	0,0,64,0,84,84,255,255,
	0,0,0,64,84,84,255,255,
	0,0,0,0,84,84,255,255,
	0,0,0,0,84,84,255,255,

	// Happy New Year
	24,24,24,24,0,0,24,24,
	0,0,0,0,0,0,0,0,
	24,24,24,24,24,0,24,24,
	0,0,0,0,0,0,0,0,
	24,24,24,24,24,0,24,24,
	119,34,20,20,8,8,8,28,
	231,98,82,82,74,74,70,230,
	119,34,34,62,34,34,34,119,
	16,112,16,16,16,16,16,124,
	56,68,4,8,16,32,68,124,
	56,68,4,24,4,4,68,56,
	12,20,36,36,126,4,4,14,
	62,32,32,60,2,2,66,60,
	14,16,32,60,34,34,34,28,
	126,66,2,4,4,8,8,8,
	60,66,66,60,66,66,66,60,
	60,66,66,66,62,2,4,120,
	
	// Heart
	0,102,153,153,129,66,36,24,
	0,0,0,0,0,0,0,0,
	0,102,153,153,129,66,36,24,
	0,0,0,0,0,0,0,0,
	0,102,153,153,129,66,36,24,
	0,0,0,0,0,0,0,0,
	0,102,153,153,129,66,36,24,
	0,102,153,153,129,64,32,16,
	0,102,153,152,128,64,32,16,
	0,96,144,144,128,64,32,16,
	0,0,128,128,128,64,32,16,
	0,0,0,0,0,64,32,16,
	
	// Cake with candles
	42,0,42,42,255,129,129,255,
	40,0,42,42,255,129,129,255,
	32,0,42,42,255,129,129,255,
	0,0,42,42,255,129,129,255,
	42,0,42,42,255,129,129,255,
	40,0,42,42,255,129,129,255,
	32,0,42,42,255,129,129,255,
	0,0,42,42,255,129,129,255,
	
};

// LEDs text to scroll array
prog_uchar ledProgramtext[] PROGMEM = {
	H,A,P,P,Y,M,O,T,H,E,R,S,D,A,Y, // 0 - 14
	H,A,P,P,Y,B,I,R,T,H,D,A,Y, // 15 - 27
	M,E,R,R,Y,C,H,R,I,S,T,M,A,S, // 28 - 41
	H,A,P,P,Y,E,A,S,T,E,R, // 42 - 52
};

// Redraw the 8x8 matrix with animation
void redraw_animation_Matrix(void) {
	leds_location = leds_location - 7; // Go to the start of the current column
	
	for (byte row = 0; row < 8; row++) {
		byte data = (unsigned char) pgm_read_byte(&ledProgram[leds_location]);
		leds_location++;
		
		for (byte column = 0; column < 8; column++) {
			if (data & (1<<column)) {
				LEDs[row][7-column] = 1;
			}
			else {
				LEDs[row][7-column] = 0;
			}
		}
	}
	
	leds_location = leds_location - 9; // Go to the end of the next column
	if (leds_location < 0) { leds_location = 0; } // Prevent underflow
}

// Append to the end 8x8 of the 8x16 matrix
void append_text_Matrix(void) {
	byte letter = (unsigned char) pgm_read_byte(&ledProgramtext[leds_location]);
	
	for (byte row = 0; row < 8; row++) {
		byte data = (unsigned char) pgm_read_byte(&ledLetters[letter][row]);
		
		for (byte column = 0; column < 8; column++) {
			if (data & (1<<column)) {
				bufferLEDs[row][15-column] = 1;
			}
			else {
				bufferLEDs[row][15-column] = 0;
			}
		}
	}
	
	leds_location++;
}

// Move the LEDs in the buffer left by one and move them to the LEDs array
void moveLEDs(void) {
	for (byte row = 0; row < 8; row++) {
		for (byte column = 15; column > 0; column--) {
			bufferLEDs[row][15-column] = bufferLEDs[row][15-(column-1)];
		}
		bufferLEDs[row][15] = 0; // End one is now off
	}
	for (byte row = 0; row < 8; row++) {
		for (byte column = 0; column < 8; column++) {
			LEDs[row][7-column] = bufferLEDs[row][7-column];
		}
	}
}

// Clear buffer LEDs
void clearLED(void) {
	for (byte row = 0; row < 8; row++) {
		for (byte column = 0; column < 8; column++) {
			bufferLEDs[row][7-column] = 0;
		}
	}
}

// Light the LEDs by going through each row/column and wait a little while
void lightLED(void) {
	for (int x = 0; x < 9; x++) {
		if (x < 8) { // Normal lighting of the LEDs
			// All columns stay low (off) and rows that have a 1 go low (on)
			PORTB = ((1 ^ LEDs[x][2])<<PB3) | ((1 ^ LEDs[x][5])<<PB2) | ((1 ^ LEDs[x][7])<<PB0);
			PORTC = ((1 ^ LEDs[x][3])<<PC3) | ((1 ^ LEDs[x][1])<<PC2) | ((1 ^ LEDs[x][0])<<PC0);
			PORTD = ((1 ^ LEDs[x][4])<<PD5) | ((1 ^ LEDs[x][6])<<PD3);
			
			// Set the column to go high (on)
			if (x == 0) { PORTD |= (1<<PD4); }
			if (x == 1) { PORTC |= (1<<PC1); }
			if (x == 2) { PORTB |= (1<<PB5); }
			if (x == 3) { PORTD |= (1<<PD7); }
			if (x == 4) { PORTB |= (1<<PB4); }
			if (x == 5) { PORTD |= (1<<PD6); }
			if (x == 6) { PORTD |= (1<<PD2); }
			if (x == 7) { PORTD |= (1<<PD1); }
		}
		else { // Our dead period to make the last row not flicker
			// Turn off all LEDs
			PORTB &= ~((1<<PB5) | (1<<PB4) | (1<<PB3) | (1<<PB2) | (1<<PB0));
			PORTC &= ~((1<<PC3) | (1<<PC2) | (1<<PC1) | (1<<PC0));
			PORTD &= ~((1<<PD7) | (1<<PD6) | (1<<PD5) | (1<<PD4) |  (1<<PD3) | (1<<PD2) | (1<<PD1));
		}
		
		_delay_ms(2); // Delay a bit so the LEDs light up for enough time
	}
}

int main(void) {
	setup();
	
	// Show battery is connected
	LEDs[0][0] = 1;
	for (int i = 0; i < 100; i++) {
		lightLED();
	}
	LEDs[0][0] = 0;
	
	
	while(1) {
		// Check we have reached the start time
		if (unix_timestamp >= triggers[trigger_location][start_time]) {
			leds_location = triggers[trigger_location][start_location]; // Set the leds start location
			
			// Increment trigger_delay until it's time to light up the LEDs
			trigger_delay += 8; // Increment by 8 seconds
			//trigger_delay += 1; // Increment by 1 second
			
			// Light up the LEDs when delay time is over
			if (trigger_delay >= triggers[trigger_location][delay_time]) {
				
				// Check if we play an animation or text
				if (triggers[trigger_location][display_type] == 0) {
					// Draw animation
					while (leds_location > triggers[trigger_location][end_location]) {
						redraw_animation_Matrix(); // Redraw the new 8x8 matrix with the next animation
						
						// Light up the LEDs a few times to fill in time
						for (int i = 0; i < triggers[trigger_location][animation_speed]; i++) {
							lightLED();
						}
					}
				}
				else {
					// Draw text
					while (leds_location <= triggers[trigger_location][end_location]) {
						
						append_text_Matrix(); // Append the 8x16 matrix with the next animation
						
						// Move the 8x16 matrix to the left 8 times
						for (int m = 0; m < 8; m++) {
							moveLEDs();
							
							// Light up the LEDs a few times to fill in time
							for (int i = 0; i < triggers[trigger_location][animation_speed]; i++) {
								lightLED();
							}
						}
					}
					
					// Clear buffer 8x16 LEDs
					clearLED();
				}
				
				trigger_delay = 0; // Reset trigger delay
			}
			
			// Check if we reached the end time, increment trigger_location to the next start time
			if (unix_timestamp >= triggers[trigger_location][end_time]) {
				
				// Check if there is a next trigger
				if (trigger_location < sizeof(triggers) / sizeof(uint32_t)) {
					trigger_location++;
					trigger_delay = 0;
				}
				else {
					// Sleep forever since there are no more triggers
					system_sleep();
				}
			}
		}
		
		// Wait until one TOSC1 cycle has elapsed
		TCCR2A = 0;
		while ((ASSR & ((1<<TCR2AUB))) != 0);
		system_sleep_pwr_save(); // Sleep for 8 seconds until the timer wakes us up
	}
}

// Overflow for 32.768KHz crystal
ISR(TIMER2_OVF_vect) {
	unix_timestamp += 8; // Increment by 8 seconds
	//unix_timestamp += 1; // Increment by 1 second
}
